######################################################################
#:ClearQuest web interface object
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::CQWeb - Subclass of CQCC::CQuest used to manage communications
with ClearQuest using the CQWeb ASP-based interface.

=head1 CLASS DESCRIPTION

This class implements the necessary methods to logon to ClearQuest,
perform queries, and modify records over the CQWeb interface. Command
strings are built up and encoded by the CCintCmd() method to identify
the user, password and database; the operation to be performed; and
the arguments needed such as the list of request numbers to have
associations created. The commands are then sent using PostCommand()
to establish a connection to the web server and handle communications.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::CQWeb
######################################################################

package CQCC::CQWeb;
use CQCC::CQuest;
@ISA = qw( CQCC::CQuest );

use strict;
sub Version { return '2.2' };

use Socket;
use CQCC::IOBase;
use CQCC::Parm;
use CQCC::Timer;

require Config;

######################################################################
#::Class Globals 
######################################################################

my @QueryResult;   # Query results buffer between methods
my $NumResultRows = 0;

######################################################################
#::method BasicCommand

=over 6

=item C<BasicCommand(keys)>

This function presets a command to the shared clauses covering
logon information, debug information and session management 
parameters.

=back

=cut

######################################################################

sub BasicCommand {
    my ($self, $db, $entity)  = @_;
    if (! defined $db) {
        $db = &GetParm("CQCC_DATABASE");
    }
    my $trace = (&GetParm("CQCC_DEBUG") > 1) ? "TRUE" : "FALSE";
    my @options = ("UID"=>$self->{USER},
                   "PWD"=>$self->{PASS},
                   "DbName"=>$db,
                   "DbgTrace"=>$trace,
                   "KeepSession"=>&GetParm("CQWEB_KEEPSESSION"));
    if (defined $entity) {
        push @options, ("EntName"=>$entity);
    }
    my $schemaRepo = &GetParm("CQCC_WEB_DATABASE_SET");
    if ($schemaRepo) {
        push @options, ("SchemaRepo"=>$schemaRepo);
    }
    return CCintCmd( @options );
}

######################################################################
#::function CCintCmd

=over 6

=item C<CCintCmd(keyList)>

Encodes pairs to a CQWeb command string and encodes strings so that
"Cmd"=>"ASS OC" produces "CCIntCmd=ASS%20OC" with &'s in between Even
though the arguments represent hash pairs they will be processed as a
list to maintain the original order

If the key in the pair is "" then it is handled as a raw "append"
request with the value being appended to the command without being
touched. This allows a precomputed string like basic_cmd to be added
on at the end.

=back

=cut

######################################################################

sub CCintCmd {
   my @pairs = @_;
   my $result="";
   while (@pairs) {
       my ($op, $value) = splice(@pairs,0, 2);
        $result .= ($result ? "&" : "") . 
                   ($op ? "CCInt$op=" . URLencode($value) :
                    $value);
   }
   return $result;
}

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=over 4

=item C<CQCC_URL>

The full URL needed to locate the ClearQuest web server. If not
provided directly it will be built up from the CQCC_SERVER and 
CQCC_SERVERROOT class parameters. 

=item C<CQCC_SERVER>

Name of the ClearQuest web server host. Used to build up the CQCC_URL
string along with CQCC_SERVERROOT.

=item C<CQCC_SERVERROOT>

Main folder on web server where ASP server will be located. 
Used to build up the CQCC_URL string along with CQCC_SERVER.

=item C<CQCC_WEB_DATABASE_SET>

Similar to CQCC_DATABASE_SET but only used by the CQWeb interface.
May or may not be set to the same value. No default value.

=item C<CQWEB_KEEPSESSION>

The CQWeb interface performance is greatly improved when the same
communications session is kept open and reused as long as it can be
(CQWEB_SESSION_TIMEOUT). This parameter enables reusing sessions and
is TRUE by default.

=item C<CQWEB_SESSION_TIMEOUT>

CQWeb communications sessions can be kept open for up to two minutes
depending on the ClearQuest release (60 seconds for 4.1, 120 seconds
for 4.2). Default is 55 seconds; bounded between 0 and 120.

=back

=back

=cut

######################################################################

sub ClassInit { 
    if (&ParmDefined("CQCC_URL")) {
        return;
    }     
    &SetParmDefaults(PCLASS=>"CQCC::CQWeb", SRC_CFG=>1);

    &DefineParm(NAME=>"CQCC_URL", SRC_EV=>1);

    &DefineParm(NAME=>"CQCC_SERVER", SRC_EV=>1);

    &DefineParm(NAME=>"CQCC_SERVERROOT", SRC_EV=>1, DEFAULT=>"cqweb" );

    &DefineParm(NAME=>"CQCC_WEB_DATABASE_SET", SRC_EV=>1);

    &DefineParm(NAME=>"CQWEB_KEEPSESSION", DEFAULT=>'TRUE',
                RANGE_LIST => ["FALSE", "TRUE"]);

    &DefineParm(NAME=>"CQWEB_SESSION_TIMEOUT", DEFAULT=>60,
                RANGE_MIN=>0, RANGE_MAX=>120);

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
calls IOClassFactory() to get an instance of the desired subclass; 
and invokes the object's SelfTest method for testing.

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;

    &ForceParm("ClearQuestInterface", "CQCC::CQWeb");

    return &CQCC::CQuest::ClassTest(@args);
}

######################################################################
#::method ClearCookie 

=over 6

=item C<ClearCookie()>

This method clears the persistent session cookie and timestamp, and 
deletes the session id file so future sessions won't attempt to use
it. 

=back

=cut

######################################################################

sub ClearCookie {
    my $self = shift;
    if ((&GetParm("CQWEB_KEEPSESSION") eq 'TRUE') && 
            -r $self->{SESSIONFILE}) {
        my $age = "(undetermined)";
        if ($self->{TIMESTAMP} > 0) {
           $age = time() - $self->{TIMESTAMP};
        } 
        &IODebugPrint("DETAILED", "ClearCookie ($age secs)\n");
        unlink $self->{SESSIONFILE};
    }
    $self->{COOKIE}='';
    $self->{COOKIE_DB}='';
    $self->{TIMESTAMP}=0;
}

######################################################################
#::method ExecCommand

=over 6

=item C<ExecCommand(cmdlist)>

This method opens up a fresh communications pipeline to ClearQuest
for each command; uses the Cookie management routines to restore or
save the session information between trigger invocations; sends the
command to ClearQuest and processes and packages the output for 
consumption by PostCommand(), its' only expected caller.

=back

=cut 

######################################################################

sub ExecCommand {
    my ($self, $cmdstring) = @_;  # Argument passed in

#:::Setup socket
    ##############################
    # Socket setup
    my $sockaddr = 'S n a4 x8';
    my ($name, $aliases, $proto) = getprotobyname('tcp');
    my ($type, $len, $thisaddr);
    ($name, $aliases, $type, $len, $thisaddr) = 
        gethostbyname($self->{LOCALHOST});
    my $rthis = pack($sockaddr, AF_INET, 0, $thisaddr);

#:::Prepare to transmit
    my $contentlen = length($cmdstring);
    # String sent to http server
    my $toserver = "POST $self->{URL} HTTP/1.0\n" .
        "Content-Length: $contentlen\n" .
        "Content-Type: application/x-www-form-urlencoded\n";

    # If no cookie then load it from session file
    # Check the time and clear any stale cookies

    $self->LoadCookie();

    if ($self->{COOKIE}) {
        $toserver .= "Cookie: $self->{COOKIE}\n";
    }

    # the \n is needed to separate the header!
    $toserver .= "\n$cmdstring";

#:::Write to the server
    # Connect directly to server indicated by URL
    my $remotehost;
    if ($self->{URL} =~ m|^http://(.*)|) {
       $remotehost = $1;
    }
    my $remoteport = 80;    # default
    my $request = "/";    # default
    ($remotehost =~ s|^([^/]+)/(.*)$|$1|) && ($request = $2);
    ($remotehost =~ s/:(\d+)$//) && ($remoteport = $1);
    &IODebugPrint("DETAILED", "Host is $remotehost:$remoteport.\n");

    # fix problem with gethostbyname() not working on decimal tcp 
    # addresses (some platforms)

    my $thataddr;
    if ($] ge "5.006") {
       $thataddr = inet_aton($remotehost);
    } else {
       my ($name, $aliases, $type, $len, $addr) = 
           gethostbyname($remotehost);
       $thataddr = $addr;
    } 

    my $there = pack($sockaddr, AF_INET, $remoteport, $thataddr);
    socket(RSVR, PF_INET, SOCK_STREAM, $proto) or 
        return("FAIL", '', '', "Can't create server socket: $!");
    connect(RSVR, $there) or return('FAIL', '', '',
        "Can't connect to $remotehost: $!");
    select(RSVR); $| = 1; select(STDOUT);

    # Remove the hostname from the URL
    if (!($toserver =~ s|http://$remotehost:$remoteport||)) {
       $toserver =~ s|http://$remotehost||;
    }

    #################################################################
    # Send the request
    &IODebugPrint("DETAILED", "SENDING TO SERVER:\n$toserver\n");
    print RSVR $toserver;

#:::Get the content
    my $length = 0;
    my $buffer;
    my $endHeader = 0;
    my $output = "";
    my ($warn,$error,$info,$result,$sessionStatus,$queryStatus);

    my $inputBuffer;
    while(read(RSVR,$inputBuffer,8*1024)) {
        $buffer .= $inputBuffer;
    }
    close RSVR;

    if ($buffer) { 
        &IODebugPrint("DETAILED", 
           "\n\n\n\n******************** Read from server:\n\n");
        my @lines = split(/^/, $buffer);
        my $line;
        foreach $line (@lines) {
            #&IODebugPrint("DETAILED", "LINE:$line\n");
            if ($endHeader && ($line =~ m/^CCInt/) ) {

                my ($key,$value) = split(/=/, $line);

                if ($key eq 'CCIntQueryRow') {
                    # if we have a query result row, add it to the 
                    # global result buffer untouched; don't know 
                    # enough about what's going on down at this
                    # level to do anything with it, except format for 
                    # debug printing.

                    $QueryResult[$NumResultRows++] = $value;

                    # this stuff is just for debug
                    my @flds = split('&',$value);
                    my $i;
                    for ($i = 0; $i <= $#flds; $i++) {
                        $flds[$i] =~ tr/+/ /;
                        $flds[$i] =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex ($1))/eg;
                    }
                    $output .= $key . '=' . $value;
                } else {
                    $value =~ tr/+/ /;
                    $value =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex ($1))/eg;
                    $output .= $key . '=' . $value;
                }
                # Check for status and messages to return to caller
                if ($line =~ m/^CCIntInfoMsg/) {
                    $info = $value;

                # SUCCESS / FAIL
                } elsif ($line =~ m/^CCIntResult/) { 
                    $result = $value;

                # ACTIVE / CLOSED
                } elsif ($line =~ m/^CCIntSessionStatus/) { 
                    $sessionStatus = $value;

                # COMPLETE / PARTIAL
                } elsif ($line =~ m/^CCIntFetchStatus/) {  
                    $queryStatus = $value;
                    &IODebugPrint("DETAILED", 
                        "QUERY STATUS: $queryStatus\n");
                } elsif ($line =~ m/^CCIntErrorMsg/) {
                    $error = $value;
                } elsif ($line =~ m/^CCIntWarnMsg/) {
                    $warn = $value;
                }
            } elsif ($line =~ m/^Set-Cookie: (.*)$/) {
                # save cookie, will throw away later if the session 
                # is gone
                $self->{COOKIE} = $1;
                &IODebugPrint("DETAILED",
                     "Cookie set: $self->{COOKIE}\n");
            } elsif ($line =~ m/^\s*$/) {
                $endHeader = 1;
            } else {
                # header stuff other than cookie
                &IODebugPrint("DETAILED","Message header: $line");
            }
        }
    } # end 'while(read(RSVR....'

#:::Finish communications - close pipe
    # don't keep the cookie if we didn't keep the session
    if ($sessionStatus =~ m/CLOSED/) { 
       $self->ClearCookie();
    } 

#:::Announce results, clear or save cookie as needed
    &IODebugPrint("DETAILED",
        "*****OUTPUT (message body) from Server:\n$output\n" .
        "*****END of Server Output\n" .
        "session status: $sessionStatus queryStat: $queryStatus " .
        "Cookie: $self->{COOKIE}\n");

    if (!$result || !$sessionStatus) {
        # if no status is read, could be bad EVs, connection down, 
        #     server misconfigured, sent bad command, etc.

        &IOMessage("ERROR", ["CQW_E_POSTCOMMAND", $self->{URL}]);
        $self->ClearCookie();
    } else {
        $self->SaveCookie();  # Save to file and reset timestamp
    } 

#:::Return value
    return ($result,$info,$warn,$error,$sessionStatus,$queryStatus);
}


######################################################################
#::method LoadCookie

=over 6

=item C<LoadCookie()>

This subroutine will load or validate the persisten session "cookie".
If not keeping session then cookie should already be empty. 
If cookie not set then read it from session file if available.
If stale cookie, then clear it and return no cookie. 

=back

=cut

######################################################################

sub LoadCookie {
    my $self = shift;

    if (&GetParm("CQWEB_KEEPSESSION") eq 'TRUE') {
       my ($age, $result);

       # NOTE: If no file, it reports a stale cookie with big timeout
       # If no cookie then restore it from file if available

       if ($self->{COOKIE} eq '' && -r $self->{SESSIONFILE}) {
          # Note - stat/$mtime is not reliable as a timestamp from 
          # the current host since it relies on the filesystem where
          # the session file is saved which may have a different clock

          &IODebugPrint("DETAILED",
                       "LoadCookie: Restoring from sessionfile\n");

          # Unless timestamp read in, will clear
          $self->{TIMESTAMP} = 0;  

          ($self->{COOKIE}, $self->{TIMESTAMP}, $self->{COOKIE_DB}) = 
              &IOFileRead($self->{SESSIONFILE},0);
       }

       # Check for stale cookie
       $age = time() - $self->{TIMESTAMP};
       if ($self->{COOKIE_DB} ne "?" &&
           $self->{COOKIE_DB} ne &GetParm("CQCC_DATABASE")) {
          &IODebugPrint("DETAILED",
              "LoadCookie - wrong cookie database: ($self->{COOKIE_DB}) != " . 
              &GetParm("CQCC_DATABASE") . "\n");
          $self->ClearCookie();
       } elsif ($age >= &GetParm("CQWEB_SESSION_TIMEOUT")) {
          &IODebugPrint("DETAILED",
              "LoadCookie - stale cookie $age secs old\n");
          $self->ClearCookie();
       } else { 
          &IODebugPrint("DETAILED", 
              "LoadCookie cookie, $age secs old\n");
       } 
    }
}

######################################################################
#::method Logon

=over 6

=item C<Logon()>

This method attempts to actively logon to the target ClearQuest
database. Once established the class will try to use the same
session id as long as possible. 

=back

=cut

######################################################################

sub Logon {
    my $self = shift;
    $self->GetLogonParms();

    &IODebugPrint("BASIC",
        "Test Mode: Trying to connect to $self->{URL}\n" . 
        "ClearQuest User is: $self->{USER}\n" . 
        "Connecting to any available DB on remote system.\n\n");

    my $t1 = TimerStart("CQWeb::Logon");
    my $trace = (&GetParm("CQCC_DEBUG") > 1) ? "TRUE" : "FALSE";
    if ($self->PostCommand("Cmd"=>"NOOP", ""=>$self->BasicCommand()))
    {
        &IODebugPrint("BASIC", "cqweb",
            "\nTest Mode SUCCESSFUL: communication succeeded with" .
            " remote server and CQ/CC integration server page.\n");
        TimerStop($t1);
        return 1;

    } else {
        &IODebugPrint("BASIC", "cqweb",
            "\nTest Mode FAILURE: communication not successful to" .
            " remote server, OR to the integration server page.\n");
        $self->ClearCookie();
        TimerStop($t1);
    }

    return 0;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::CQWeb()>

This method initializes the object's attributes including the 
URL, name of the session file to store session information between
trigger invocations and the session id (cookie) used. 

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self;

   # CCperl doesn't support SUPER so use direct reference to parent
   $self = new CQCC::CQuest();

   $self->{CLASSNAME} = "CQCC::CQWeb";
   $self->{URL} =           # URL to CQ web database
       &GetParm("CQCC_URL");

   # Don't rely on session file age, use embedded timestamp from client
   $self->{SESSIONFILE} =   # File containing sessionid, ~/.cqsession
       &IOResourceManager(".cqsession", 1, &GetParm("CM_CQSESSION_STALE_MINS")); 
   $self->{COOKIE} = "";    # CQ sessionid returned by web server
                            # Stored into ~/.cqsession file 
   $self->{COOKIE_DB} = "?"; # CQ database associated with cookie

   my $hostname = "";       # On PC, uninitialized value works ok
                            # On UNIX, ask OS what hostname is
   if ($Config::Config{osname} ne "MSWin32") {
       $hostname = `uname -n`;
       chop($hostname);
   }
   $self->{LOCALHOST} = $hostname;  # Local hostname that gethostbyname uses

   if (! defined $self->{URL}) {
       if (&GetParm("CQCC_SERVER") && &GetParm("CQCC_SERVERROOT")) {
           $self->{URL} = "http://" . &GetParm("CQCC_SERVER") . 
               "\/" . &GetParm("CQCC_SERVERROOT") .
               "/ccintegration/default.asp";
       }
   }

   # Error if still not defined
   if (! defined $self->{URL}) {
       &IOMessage("FATAL", ["CQW_E_INVALID_SERVER"]);
   }

   bless($self, $class);
   return $self;
}

######################################################################
#::method PostCommand

=over 6

=item C<PostCommand($cmd)>

This method completes preparation of the command being sent, sends it
to ClearQuest using the ExecCommand() method, then interprets the
result codes returned from ClearQuest.

The return values are (0) - Failed; (1) Succeeded; and 2(Succeeded
and more data is available. 

=back

=cut

######################################################################

sub PostCommand {
    my ($self, @cmdArgs) = @_;

    my $cmd = CCintCmd( @cmdArgs );
    my $result = 1;

    my $t1 = &TimerStart("CQWeb::PostCommand $cmdArgs[1]");
    &IODebugPrint("DETAILED", "PostCommand: $cmd\n");
    
    my ($stat, $info, $warn, $error, $sessionStat, $queryStat) = 
        $self->ExecCommand($cmd);

    chomp $stat;

    &IODebugPrint("DETAILED",
        "\n\nASP returned status: $stat\n" .
        "ASP returned msgs:\n    info: $info\n    warn: $warn\n" .
        "    error: $error\n");

    if ($error =~ "Database name is not valid") {
        &IOMessage("ERROR",["CQW_E_DATABASE_INVALID", $error]);
    }

    if ($error =~ /Code Page Mismatch/g) {
        &IOMessage("FATAL", [ "CQP_E_CODE_PAGE_MISMATCH"]);
    }

    if ($stat =~ /FAIL/) {
        # operation failed - should report CCIntCmd name
        &IOMessage("ERROR",["CQW_E_OP_FAILED", $warn, $error]);
        $result = 0;
    } elsif ($queryStat =~ m/PARTIAL/ && $sessionStat =~ m/ACTIVE/) {
        $result = 2;
    }

    &TimerStop($t1);
    return $result;
}

######################################################################
#::method ProcessBatchOps

=over 6

=item C<ProcessBatchOps(hash)>

This method processes the hash lists built up when the batch information
was processed from a file or other means. 

=back

=cut

######################################################################

sub ProcessBatchOps {
    my ($self,$cqHash) = @_;
    my %cqHash = %{$cqHash};
    my (%volHash, %addHash, %delHash, $key, $cqKey);
    my $result = 1;

    # Break up the {defect}->({path}->(op,vobid,oid,oldOID),...) order
    # into several hash tables for easier processing by CQWeb
    # Key based on "database,entity,pathname" to clump similar db-entity
    # sets together. 
    #    volHash:  {key}->"path,vobid,oid,oldOID"
    #    addHash:  {key}->"defect,defect,..."
    #    delHash:  {key}->"defect,defect,..."

    for $cqKey (sort keys %cqHash) {
        my ($entity,$dbID) = split(":",$cqKey);
        my $db = substr($dbID,0,length($dbID)-8);
        my $id = substr($dbID,-8);
        my %opHash = %{$cqHash{$cqKey}};
        my $path; 

        for $path (sort keys %opHash) {
            my ($op,$vobid,$oid,$oldOID) = split(",", $opHash{$path});
            $key = join(",", $db, $entity, $path);
            if (! defined $volHash{$key}) {
                $volHash{$key} = join(",",$path,$vobid,$oid,$oldOID);
                if ($op eq "<+>") {
                    $addHash{$key} = $id;
                } else {
                    $delHash{$key} = $id;
                }
            }
            else {
               if ($op eq "<+>") {
                   $addHash{$key} = $addHash{$key} . "," . $id;
               } else {
                   $delHash{$key} = $delHash{$key} . "," . $id;
               }
            }
        }
    }
    my $addAssocs = new CQCC::AssocList();
    my $delAssocs = new CQCC::AssocList();
    for $key (sort keys %volHash) {
         my ($db, $entity, $path) = split(",", $key);
         my $voList = $volHash{$key};
         $self->SetDBEntity($db,$entity);
         $addAssocs->ProcessBugString($entity,$addHash{$key});
         $delAssocs->ProcessBugString($entity,$delHash{$key});
         my $r = $self->EditAssocs($voList, $addAssocs, $delAssocs);
         if ($r == 0) {
             $result = 0;
             return $result;
         }
    }
    return $result;
}

######################################################################
#::method SaveCookie 

=over 6

=item C<SaveCookie()>

This method will save the persistent session cookie along with a 
timestamp to ensure the session information is not used after it
has expired. 

=back

=cut

######################################################################

sub SaveCookie {
    my $self = shift;

    if ((&GetParm("CQWEB_KEEPSESSION") eq 'TRUE') && 
            ($self->{COOKIE} ne '')) {
        my $result;
        &IODebugPrint("DETAILED", 
            "SaveCookie cookie into sessionfile\n");
        $self->{TIMESTAMP} = time();
        $self->{COOKIE_DB} = &GetParm("CQCC_DATABASE");
        &IOFileWrite($self->{SESSIONFILE}, 
            $self->{COOKIE}, $self->{TIMESTAMP}, $self->{COOKIE_DB});
        # If unable to open sessionfile for writing, won't be able to
        # reuse sessionid on next trigger but will be ok on this one 
        # so don't report an error
    }
}

######################################################################
#::function URLencode

=over 6

=item C<URLencode($value)>

Internal function used to encode the commands and arguments being
sent of the pipe. Originally from CGI.PM.

=back

=cut

######################################################################

sub URLencode 
{
    my($toencode) = @_;
    # /eg means right of substitution is an Expression, Global replace
    # Looks for anything EXCEPT a-z, A-Z, 0-9, _, -, and .
    # Encodes the offending chars to %format

    $toencode =~ s/([^a-zA-Z0-9_\-.])/uc sprintf("%%%02x",ord($1))/eg;
    return $toencode;
}

######################################################################
#::method _EditAssocs

=over 6

=item C<_EditAssocs($dbname,$entity,$VOList,$addAssocs,$delAssocs)>

This internal method takes information about the ClearCase version
object being associated to and two AssocLists, one listing the new
associations to be made, and the other, the list of associations to be
deleted.  If the system fails to successfully make the new additions,
they are all deleted to prepare for the trigger failing gracefully.

It must be called thru the CQuest method, EditAssocs which will 
break up the association lists into unique database/entity 
combinations so that this method only handles one combination at a 
time. 

Returns 1 for success, 0 for failure.

=back

=cut

######################################################################

sub _EditAssocs {
    my ($self,$dbName,$entity,$VOList,$addAssocs,$delAssocs) = @_;

    my ($cqname, $vobid, $oid, $oldOID) = split(",", $VOList);
    my $basicCmd = $self->BasicCommand($dbName, $entity);
    my $VOstr = URLencode("$vobid\001$oid\001$cqname");
    my $addIDs = (defined $addAssocs) ? 
        $addAssocs->GetBugList(undef,'&') : undef;
    my $delIDs = (defined $delAssocs) ?
        $delAssocs->GetBugList(undef,'&') : undef;
    my $result = 1;
  
    if ($delIDs) {
        my $VOstr = URLencode("$vobid\001$oldOID\001" . 
            &GetParm("CLEARCASE_PN"));
        $result = $self->PostCommand("Cmd"=>"DISASSOCIATE", 
            "BugIDs"=>$delIDs, "VOList"=>$VOstr, ""=>$basicCmd);
    }

    if ($addIDs) {
        if (! $self->PostCommand("Cmd"=>"ASSOCIATE", 
            "BugIDs"=>$addIDs, "VOList"=>$VOstr, ""=>$basicCmd))
        {
            # association failed; attempt to DISassociate the same 
            # list in order to clean up as much as possible.

            $self->PostCommand("Cmd"=>"DISASSOCIATE", 
                "BugIDs"=>$addIDs, "VOList"=>$VOstr, ""=>$basicCmd);

            $result = 0;
        }
    }

    return $result;
}

######################################################################
#::method _PerformQuery 

=over 6

=item C<_PerformQuery(db,entity,schema,query,results,results_format,
returnType, queryType)>

This internal method is called by PerformQuery to actually query the
ClearQuest database for records matching the query description.

Queries are the only type of request where we may have to generate
more than one command to get the entire result back; if the query
status returns 'PARTIAL' we need to send a 'FETCH' command to obtain
more result rows. Also, session status remains ACTIVE if more rows are
to be fetched, regardless of the 'KeepSession' setting.

=back

=cut

######################################################################

sub _PerformQuery {
    my ($self, $dbname, $entity, $schema, $query, $resultsSet,
        $resultsFormat, $returnType, $queryType) = @_;

    if ($queryType eq "CQ") { 
        &IOMessage("ERROR", ["CQ_E_OP_UNSUPPORTED", 
                             "CQWeb", 
                             "PerformQuery with Named queries"]);
        return;
    }

    my @clauseList = @{ $query };  # Must be an array reference
    my @filters;

    # Postprocess the query to provide URL encoding of the query
    for (@clauseList) {
        my ($subject, $verb, $args) = split(" ", $_, 3);
        my $filter = URLencode($subject) . "\001" . $verb;
        for (split(",", $args)) {
            $filter .= ("\001" . URLencode($_));
       }
       push @filters, $filter;
    }

    undef @QueryResult;         # Results from raw Query operation 
    $NumResultRows = 0;

    my $resultsFields = join("&", split(",", $resultsSet));    
   
    # Execute the Query command then Fetch the results
    my $result = $self->PostCommand("Cmd"=>"QUERY",
        "QryFlds" => $resultsFields,
        "QryFltrs" => URLencode(join('&',@filters)),
        ""=> $self->BasicCommand($dbname, $entity));
    my $trace = (&GetParm("CQCC_DEBUG") > 1) ? "TRUE" : "FALSE";

    if (! $result) {
        return ();
    }

    while ($result == 2) {
        $result = 
            $self->PostCommand("Cmd"=>"FETCH",
                               ""=>$self->BasicCommand());
        if (! $result) {
            return ();
        }
    }

    my @bugdesc;
    my $encodedbug;
    # query results come back in a global array, still url-encoded. 
    # Decode and format them.

    my $n;

    # Include column labels as first record in return
    if ($returnType eq "ARRAY") { 
        my (@fld,$i);
        my @fieldLabels = split(",", $resultsSet);
        for ($i = 0; $i <= $#fieldLabels; $i++) {
            $fld[$i] = $fieldLabels[$i];
        }
        $bugdesc[$n++] = \@fld;
    }

    foreach $encodedbug (@QueryResult) {
        # get rid of NT-style cr/lf combo sent by NT server
        chop $encodedbug;
        chop $encodedbug;
        my @fld = split('&',$encodedbug);
        my $i;
        for ($i = 0; $i <= $#fld; $i++) {
            $fld[$i] =~ tr/+/ /;
            $fld[$i] =~ 
                s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex ($1))/eg;
        }
        if ($returnType eq "STRING") { 
            $bugdesc[$n++] = sprintf($resultsFormat, @fld);
        }
        else {
            my @tempFld = @fld;
            $bugdesc[$n++] = \@tempFld;
        }
    }

    if ($returnType eq "STRING") {
        return sort @bugdesc;
    }
    return @bugdesc;
}

######################################################################
#::Tail
######################################################################

1;
